@@ -37,6 +37,7 @@
 //! ```
 use crate::{
     paths::Location,
+    thread::ThreadBound,
     types::{JsonType, JsonTypeSet},
 };
 use serde_json::{Map, Number, Value};
@@ -78,7 +79,15 @@ pub struct ValidationError<'a> {
 ///     }
 /// }
 /// ```
-pub type ErrorIterator<'a> = Box<dyn Iterator<Item = ValidationError<'a>> + Sync + Send + 'a>;
+#[doc(hidden)]
+pub trait ValidationErrorIterator<'a>: Iterator<Item = ValidationError<'a>> + ThreadBound {}
+
+impl<'a, T> ValidationErrorIterator<'a> for T where
+    T: Iterator<Item = ValidationError<'a>> + ThreadBound
+{
+}
+
+pub type ErrorIterator<'a> = Box<dyn ValidationErrorIterator<'a> + 'a>;
 
 // Empty iterator means no error happened
 pub(crate) fn no_error<'a>() -> ErrorIterator<'a> {
@@ -142,7 +151,12 @@ pub enum ValidationErrorKind {
     /// Not enough properties in an object.
     MinProperties { limit: u64 },
     /// When some number is not a multiple of another number.
-    MultipleOf { multiple_of: f64 },
+    MultipleOf {
+        #[cfg(feature = "arbitrary-precision")]
+        multiple_of: Value,
+        #[cfg(not(feature = "arbitrary-precision"))]
+        multiple_of: f64,
+    },
     /// Negated schema failed validation.
     Not { schema: Value },
     /// The given schema is valid under more than one of the schemas listed in the 'oneOf' keyword.
@@ -173,6 +187,49 @@ pub enum ValidationErrorKind {
     Referencing(referencing::Error),
 }
 
+impl ValidationErrorKind {
+    pub(crate) fn keyword(&self) -> &'static str {
+        match self {
+            ValidationErrorKind::AdditionalItems { .. } => "additionalItems",
+            ValidationErrorKind::AdditionalProperties { .. } => "additionalProperties",
+            ValidationErrorKind::AnyOf { .. } => "anyOf",
+            ValidationErrorKind::BacktrackLimitExceeded { .. }
+            | ValidationErrorKind::Pattern { .. } => "pattern",
+            ValidationErrorKind::Constant { .. } => "const",
+            ValidationErrorKind::Contains => "contains",
+            ValidationErrorKind::ContentEncoding { .. } | ValidationErrorKind::FromUtf8 { .. } => {
+                "contentEncoding"
+            }
+            ValidationErrorKind::ContentMediaType { .. } => "contentMediaType",
+            ValidationErrorKind::Custom { .. } => "custom",
+            ValidationErrorKind::Enum { .. } => "enum",
+            ValidationErrorKind::ExclusiveMaximum { .. } => "exclusiveMaximum",
+            ValidationErrorKind::ExclusiveMinimum { .. } => "exclusiveMinimum",
+            ValidationErrorKind::FalseSchema => "falseSchema",
+            ValidationErrorKind::Format { .. } => "format",
+            ValidationErrorKind::MaxItems { .. } => "maxItems",
+            ValidationErrorKind::Maximum { .. } => "maximum",
+            ValidationErrorKind::MaxLength { .. } => "maxLength",
+            ValidationErrorKind::MaxProperties { .. } => "maxProperties",
+            ValidationErrorKind::MinItems { .. } => "minItems",
+            ValidationErrorKind::Minimum { .. } => "minimum",
+            ValidationErrorKind::MinLength { .. } => "minLength",
+            ValidationErrorKind::MinProperties { .. } => "minProperties",
+            ValidationErrorKind::MultipleOf { .. } => "multipleOf",
+            ValidationErrorKind::Not { .. } => "not",
+            ValidationErrorKind::OneOfMultipleValid { .. }
+            | ValidationErrorKind::OneOfNotValid { .. } => "oneOf",
+            ValidationErrorKind::PropertyNames { .. } => "propertyNames",
+            ValidationErrorKind::Required { .. } => "required",
+            ValidationErrorKind::Type { .. } => "type",
+            ValidationErrorKind::UnevaluatedItems { .. } => "unevaluatedItems",
+            ValidationErrorKind::UnevaluatedProperties { .. } => "unevaluatedProperties",
+            ValidationErrorKind::UniqueItems => "uniqueItems",
+            ValidationErrorKind::Referencing(_) => "$ref",
+        }
+    }
+}
+
 #[derive(Debug)]
 #[allow(missing_docs)]
 pub enum TypeKind {
@@ -578,6 +635,22 @@ impl<'a> ValidationError<'a> {
             schema_path: location,
         }
     }
+    #[cfg(feature = "arbitrary-precision")]
+    pub(crate) fn multiple_of(
+        location: Location,
+        instance_path: Location,
+        instance: &'a Value,
+        multiple_of: Value,
+    ) -> ValidationError<'a> {
+        ValidationError {
+            instance_path,
+            instance: Cow::Borrowed(instance),
+            kind: ValidationErrorKind::MultipleOf { multiple_of },
+            schema_path: location,
+        }
+    }
+
+    #[cfg(not(feature = "arbitrary-precision"))]
     pub(crate) const fn multiple_of(
         location: Location,
         instance_path: Location,
