@@ -9,10 +9,10 @@
 use crate::{
     compiler,
     error::{no_error, ErrorIterator, ValidationError},
+    evaluation::{format_schema_location, Annotations, ErrorDescription, EvaluationNode},
     keywords::CompilationResult,
     node::SchemaNode,
     options::PatternEngineOptions,
-    output::{Annotations, BasicOutput, OutputUnit},
     paths::{LazyLocation, Location},
     properties::{
         are_properties_valid, compile_big_map, compile_dynamic_prop_map_validator,
@@ -21,10 +21,11 @@ use crate::{
     },
     regex::RegexEngine,
     types::JsonType,
-    validator::{PartialApplication, Validate},
+    validator::{EvaluationResult, Validate},
 };
 use referencing::Uri;
 use serde_json::{Map, Value};
+use std::sync::Arc;
 
 macro_rules! is_valid {
     ($node:expr, $value:ident) => {{
@@ -130,20 +131,23 @@ impl Validate for AdditionalPropertiesValidator {
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
-            let mut matched_props = Vec::with_capacity(item.len());
-            let mut output = BasicOutput::default();
+            let mut children = Vec::with_capacity(item.len());
             for (name, value) in item {
                 let path = location.push(name.as_str());
-                output += self.node.apply_rooted(value, &path);
-                matched_props.push(name.clone());
+                children.push(self.node.evaluate_instance(value, &path));
             }
-            let mut result: PartialApplication = output.into();
-            result.annotate(Value::from(matched_props).into());
+            let mut result = EvaluationResult::from_children(children);
+            let annotated_props = item
+                .keys()
+                .cloned()
+                .map(serde_json::Value::String)
+                .collect();
+            result.annotate(Annotations::new(serde_json::Value::Array(annotated_props)));
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -301,33 +305,32 @@ impl<M: PropertiesValidatorsMap> Validate for AdditionalPropertiesNotEmptyFalseV
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
             let mut unexpected = Vec::with_capacity(item.len());
-            let mut output = BasicOutput::default();
+            let mut children = Vec::with_capacity(item.len());
             for (property, value) in item {
                 if let Some((_name, node)) = self.properties.get_key_validator(property) {
                     let path = location.push(property.as_str());
-                    output += node.apply_rooted(value, &path);
+                    children.push(node.evaluate_instance(value, &path));
                 } else {
                     unexpected.push(property.clone());
                 }
             }
-            let mut result: PartialApplication = output.into();
+            let mut result = EvaluationResult::from_children(children);
             if !unexpected.is_empty() {
-                result.mark_errored(
-                    ValidationError::additional_properties(
+                result.mark_errored(ErrorDescription::from_validation_error(
+                    &ValidationError::additional_properties(
                         self.location.clone(),
                         location.into(),
                         instance,
                         unexpected,
-                    )
-                    .into(),
-                );
+                    ),
+                ));
             }
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -419,38 +422,39 @@ impl<M: PropertiesValidatorsMap> Validate for AdditionalPropertiesNotEmptyValida
     ) -> Result<(), ValidationError<'i>> {
         if let Value::Object(props) = instance {
             for (property, instance) in props {
+                let property_location = location.push(property);
                 if let Some(validator) = self.properties.get_validator(property) {
-                    validator.validate(instance, &location.push(property))?;
+                    validator.validate(instance, &property_location)?;
                 } else {
-                    self.node.validate(instance, &location.push(property))?;
+                    self.node.validate(instance, &property_location)?;
                 }
             }
         }
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(map) = instance {
             let mut matched_propnames = Vec::with_capacity(map.len());
-            let mut output = BasicOutput::default();
+            let mut children = Vec::with_capacity(map.len());
             for (property, value) in map {
                 let path = location.push(property.as_str());
                 if let Some((_name, property_validators)) =
                     self.properties.get_key_validator(property)
                 {
-                    output += property_validators.apply_rooted(value, &path);
+                    children.push(property_validators.evaluate_instance(value, &path));
                 } else {
-                    output += self.node.apply_rooted(value, &path);
+                    children.push(self.node.evaluate_instance(value, &path));
                     matched_propnames.push(property.clone());
                 }
             }
-            let mut result: PartialApplication = output.into();
+            let mut result = EvaluationResult::from_children(children);
             if !matched_propnames.is_empty() {
-                result.annotate(Value::from(matched_propnames).into());
+                result.annotate(Annotations::new(Value::from(matched_propnames)));
             }
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -484,7 +488,7 @@ pub(crate) struct AdditionalPropertiesWithPatternsValidator<R> {
     /// "additionalProperties" as it's path. However, we need to produce annotations which have the
     /// patternProperties keyword as their path so we store the paths here.
     pattern_keyword_path: Location,
-    pattern_keyword_absolute_location: Option<Uri<String>>,
+    pattern_keyword_absolute_location: Option<Arc<Uri<String>>>,
 }
 
 impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsValidator<R> {
@@ -537,26 +541,27 @@ impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsValidator<R> {
     ) -> Result<(), ValidationError<'i>> {
         if let Value::Object(item) = instance {
             for (property, value) in item {
+                let property_location = location.push(property);
                 let mut has_match = false;
                 for (re, node) in &self.patterns {
                     if re.is_match(property).unwrap_or(false) {
                         has_match = true;
-                        node.validate(value, &location.push(property))?;
+                        node.validate(value, &property_location)?;
                     }
                 }
                 if !has_match {
-                    self.node.validate(value, &location.push(property))?;
+                    self.node.validate(value, &property_location)?;
                 }
             }
         }
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
-            let mut output = BasicOutput::default();
             let mut pattern_matched_propnames = Vec::with_capacity(item.len());
             let mut additional_matched_propnames = Vec::with_capacity(item.len());
+            let mut children = Vec::with_capacity(item.len());
             for (property, value) in item {
                 let path = location.push(property.as_str());
                 let mut has_match = false;
@@ -564,30 +569,36 @@ impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsValidator<R> {
                     if pattern.is_match(property).unwrap_or(false) {
                         has_match = true;
                         pattern_matched_propnames.push(property.clone());
-                        output += node.apply_rooted(value, &path);
+                        children.push(node.evaluate_instance(value, &path));
                     }
                 }
                 if !has_match {
                     additional_matched_propnames.push(property.clone());
-                    output += self.node.apply_rooted(value, &path);
+                    children.push(self.node.evaluate_instance(value, &path));
                 }
             }
             if !pattern_matched_propnames.is_empty() {
-                output += OutputUnit::<Annotations<'_>>::annotations(
+                let annotation = Annotations::new(Value::from(pattern_matched_propnames));
+                let schema_location = format_schema_location(
+                    &self.pattern_keyword_path,
+                    self.pattern_keyword_absolute_location.as_ref(),
+                );
+                children.push(EvaluationNode::valid(
                     self.pattern_keyword_path.clone(),
-                    location.into(),
                     self.pattern_keyword_absolute_location.clone(),
-                    Value::from(pattern_matched_propnames).into(),
-                )
-                .into();
+                    schema_location,
+                    location.into(),
+                    Some(annotation),
+                    Vec::new(),
+                ));
             }
-            let mut result: PartialApplication = output.into();
+            let mut result = EvaluationResult::from_children(children);
             if !additional_matched_propnames.is_empty() {
-                result.annotate(Value::from(additional_matched_propnames).into());
+                result.annotate(Annotations::new(Value::from(additional_matched_propnames)));
             }
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -617,7 +628,7 @@ pub(crate) struct AdditionalPropertiesWithPatternsFalseValidator<R> {
     patterns: Vec<(R, SchemaNode)>,
     location: Location,
     pattern_keyword_path: Location,
-    pattern_keyword_absolute_location: Option<Uri<String>>,
+    pattern_keyword_absolute_location: Option<Arc<Uri<String>>>,
 }
 
 impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsFalseValidator<R> {
@@ -671,11 +682,12 @@ impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsFalseValidator
     ) -> Result<(), ValidationError<'i>> {
         if let Value::Object(item) = instance {
             for (property, value) in item {
+                let property_location = location.push(property);
                 let mut has_match = false;
                 for (re, node) in &self.patterns {
                     if re.is_match(property).unwrap_or(false) {
                         has_match = true;
-                        node.validate(value, &location.push(property))?;
+                        node.validate(value, &property_location)?;
                     }
                 }
                 if !has_match {
@@ -691,11 +703,11 @@ impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsFalseValidator
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
-            let mut output = BasicOutput::default();
             let mut unexpected = Vec::with_capacity(item.len());
             let mut pattern_matched_props = Vec::with_capacity(item.len());
+            let mut children = Vec::with_capacity(item.len());
             for (property, value) in item {
                 let path = location.push(property.as_str());
                 let mut has_match = false;
@@ -703,7 +715,7 @@ impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsFalseValidator
                     if pattern.is_match(property).unwrap_or(false) {
                         has_match = true;
                         pattern_matched_props.push(property.clone());
-                        output += node.apply_rooted(value, &path);
+                        children.push(node.evaluate_instance(value, &path));
                     }
                 }
                 if !has_match {
@@ -711,29 +723,34 @@ impl<R: RegexEngine> Validate for AdditionalPropertiesWithPatternsFalseValidator
                 }
             }
             if !pattern_matched_props.is_empty() {
-                output += OutputUnit::<Annotations<'_>>::annotations(
+                let annotation = Annotations::new(Value::from(pattern_matched_props));
+                let schema_location = format_schema_location(
+                    &self.pattern_keyword_path,
+                    self.pattern_keyword_absolute_location.as_ref(),
+                );
+                children.push(EvaluationNode::valid(
                     self.pattern_keyword_path.clone(),
-                    location.into(),
                     self.pattern_keyword_absolute_location.clone(),
-                    Value::from(pattern_matched_props).into(),
-                )
-                .into();
+                    schema_location,
+                    location.into(),
+                    Some(annotation),
+                    Vec::new(),
+                ));
             }
-            let mut result: PartialApplication = output.into();
+            let mut result = EvaluationResult::from_children(children);
             if !unexpected.is_empty() {
-                result.mark_errored(
-                    ValidationError::additional_properties(
+                result.mark_errored(ErrorDescription::from_validation_error(
+                    &ValidationError::additional_properties(
                         self.location.clone(),
                         location.into(),
                         instance,
                         unexpected,
-                    )
-                    .into(),
-                );
+                    ),
+                ));
             }
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -851,23 +868,25 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
         if let Value::Object(item) = instance {
             for (property, value) in item {
                 if let Some((name, node)) = self.properties.get_key_validator(property) {
-                    node.validate(value, &location.push(name))?;
+                    let name_location = location.push(name);
+                    node.validate(value, &name_location)?;
                     for (re, node) in &self.patterns {
                         if re.is_match(property).unwrap_or(false) {
-                            node.validate(value, &location.push(name))?;
+                            node.validate(value, &name_location)?;
                         }
                     }
                 } else {
+                    let property_location = location.push(property);
                     let mut has_match = false;
                     for (re, node) in &self.patterns {
                         if re.is_match(property).unwrap_or(false) {
                             has_match = true;
-                            node.validate(value, &location.push(property))?;
+                            node.validate(value, &property_location)?;
                         }
                     }
 
                     if !has_match {
-                        self.node.validate(value, &location.push(property))?;
+                        self.node.validate(value, &property_location)?;
                     }
                 }
             }
@@ -875,17 +894,17 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
-            let mut output = BasicOutput::default();
             let mut additional_matches = Vec::with_capacity(item.len());
+            let mut children = Vec::with_capacity(item.len());
             for (property, value) in item {
                 let path = location.push(property.as_str());
                 if let Some((_name, node)) = self.properties.get_key_validator(property) {
-                    output += node.apply_rooted(value, &path);
+                    children.push(node.evaluate_instance(value, &path));
                     for (pattern, node) in &self.patterns {
                         if pattern.is_match(property).unwrap_or(false) {
-                            output += node.apply_rooted(value, &path);
+                            children.push(node.evaluate_instance(value, &path));
                         }
                     }
                 } else {
@@ -893,20 +912,20 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
                     for (pattern, node) in &self.patterns {
                         if pattern.is_match(property).unwrap_or(false) {
                             has_match = true;
-                            output += node.apply_rooted(value, &path);
+                            children.push(node.evaluate_instance(value, &path));
                         }
                     }
                     if !has_match {
                         additional_matches.push(property.clone());
-                        output += self.node.apply_rooted(value, &path);
+                        children.push(self.node.evaluate_instance(value, &path));
                     }
                 }
             }
-            let mut result: PartialApplication = output.into();
-            result.annotate(Value::from(additional_matches).into());
+            let mut result = EvaluationResult::from_children(children);
+            result.annotate(Annotations::new(Value::from(additional_matches)));
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -1026,18 +1045,20 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
             // No properties are allowed, except ones defined in `properties` or `patternProperties`
             for (property, value) in item {
                 if let Some((name, node)) = self.properties.get_key_validator(property) {
-                    node.validate(value, &location.push(name))?;
+                    let name_location = location.push(name);
+                    node.validate(value, &name_location)?;
                     for (re, node) in &self.patterns {
                         if re.is_match(property).unwrap_or(false) {
-                            node.validate(value, &location.push(name))?;
+                            node.validate(value, &name_location)?;
                         }
                     }
                 } else {
+                    let property_location = location.push(property);
                     let mut has_match = false;
                     for (re, node) in &self.patterns {
                         if re.is_match(property).unwrap_or(false) {
                             has_match = true;
-                            node.validate(value, &location.push(property))?;
+                            node.validate(value, &property_location)?;
                         }
                     }
                     if !has_match {
@@ -1054,18 +1075,18 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
-            let mut output = BasicOutput::default();
             let mut unexpected = vec![];
+            let mut children = Vec::with_capacity(item.len());
             // No properties are allowed, except ones defined in `properties` or `patternProperties`
             for (property, value) in item {
                 let path = location.push(property.as_str());
                 if let Some((_name, node)) = self.properties.get_key_validator(property) {
-                    output += node.apply_rooted(value, &path);
+                    children.push(node.evaluate_instance(value, &path));
                     for (pattern, node) in &self.patterns {
                         if pattern.is_match(property).unwrap_or(false) {
-                            output += node.apply_rooted(value, &path);
+                            children.push(node.evaluate_instance(value, &path));
                         }
                     }
                 } else {
@@ -1073,7 +1094,7 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
                     for (pattern, node) in &self.patterns {
                         if pattern.is_match(property).unwrap_or(false) {
                             has_match = true;
-                            output += node.apply_rooted(value, &path);
+                            children.push(node.evaluate_instance(value, &path));
                         }
                     }
                     if !has_match {
@@ -1081,21 +1102,20 @@ impl<M: PropertiesValidatorsMap, R: RegexEngine> Validate
                     }
                 }
             }
-            let mut result: PartialApplication = output.into();
+            let mut result = EvaluationResult::from_children(children);
             if !unexpected.is_empty() {
-                result.mark_errored(
-                    ValidationError::additional_properties(
+                result.mark_errored(ErrorDescription::from_validation_error(
+                    &ValidationError::additional_properties(
                         self.location.clone(),
                         location.into(),
                         instance,
                         unexpected,
-                    )
-                    .into(),
-                );
+                    ),
+                ));
             }
             result
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
@@ -1177,18 +1197,8 @@ pub(crate) fn compile<'a>(
         if let Value::Object(obj) = patterns {
             // Compile all patterns & their validators to avoid doing work in the `patternProperties` validator
             match ctx.config().pattern_options() {
-                PatternEngineOptions::FancyRegex {
-                    backtrack_limit,
-                    size_limit,
-                    dfa_size_limit,
-                } => {
-                    let patterns = match compile_fancy_regex_patterns(
-                        ctx,
-                        obj,
-                        backtrack_limit,
-                        size_limit,
-                        dfa_size_limit,
-                    ) {
+                PatternEngineOptions::FancyRegex { .. } => {
+                    let patterns = match compile_fancy_regex_patterns(ctx, obj) {
                         Ok(patterns) => patterns,
                         Err(error) => return Some(Err(error)),
                     };
@@ -1254,15 +1264,11 @@ pub(crate) fn compile<'a>(
                         }
                     }
                 }
-                PatternEngineOptions::Regex {
-                    size_limit,
-                    dfa_size_limit,
-                } => {
-                    let patterns =
-                        match compile_regex_patterns(ctx, obj, size_limit, dfa_size_limit) {
-                            Ok(patterns) => patterns,
-                            Err(error) => return Some(Err(error)),
-                        };
+                PatternEngineOptions::Regex { .. } => {
+                    let patterns = match compile_regex_patterns(ctx, obj) {
+                        Ok(patterns) => patterns,
+                        Err(error) => return Some(Err(error)),
+                    };
                     match schema {
                         Value::Bool(true) => None, // "additionalProperties" are "true" by default
                         Value::Bool(false) => {
