@@ -4,7 +4,7 @@ use crate::{
     keywords::CompilationResult,
     node::SchemaNode,
     paths::{LazyLocation, Location},
-    validator::{PartialApplication, Validate},
+    validator::{EvaluationResult, Validate},
 };
 use serde_json::{Map, Value};
 
@@ -29,7 +29,7 @@ impl Validate for PropertyNamesObjectValidator {
             let errors: Vec<_> = item
                 .keys()
                 .flat_map(move |key| {
-                    let wrapper = Value::String(key.to_string());
+                    let wrapper = Value::String(key.clone());
                     let errors: Vec<_> = self
                         .node
                         .iter_errors(&wrapper, location)
@@ -54,7 +54,7 @@ impl Validate for PropertyNamesObjectValidator {
     fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Object(item) = &instance {
             item.keys().all(move |key| {
-                let wrapper = Value::String(key.to_string());
+                let wrapper = Value::String(key.clone());
                 self.node.is_valid(&wrapper)
             })
         } else {
@@ -69,7 +69,7 @@ impl Validate for PropertyNamesObjectValidator {
     ) -> Result<(), ValidationError<'i>> {
         if let Value::Object(item) = &instance {
             for key in item.keys() {
-                let wrapper = Value::String(key.to_string());
+                let wrapper = Value::String(key.clone());
                 match self.node.validate(&wrapper, location) {
                     Ok(()) => {}
                     Err(error) => {
@@ -86,16 +86,16 @@ impl Validate for PropertyNamesObjectValidator {
         Ok(())
     }
 
-    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
+    fn evaluate(&self, instance: &Value, location: &LazyLocation) -> EvaluationResult {
         if let Value::Object(item) = instance {
-            item.keys()
-                .map(|key| {
-                    let wrapper = Value::String(key.to_string());
-                    self.node.apply_rooted(&wrapper, location)
-                })
-                .collect()
+            let mut children = Vec::with_capacity(item.len());
+            for key in item.keys() {
+                let wrapper = Value::String(key.clone());
+                children.push(self.node.evaluate_instance(&wrapper, location));
+            }
+            EvaluationResult::from_children(children)
         } else {
-            PartialApplication::valid_empty()
+            EvaluationResult::valid_empty()
         }
     }
 }
